@@ -19,11 +19,13 @@
 
 use glob::glob;
 use same_file::is_same_file;
-use serde::Serialize;
-use std::collections::HashMap;
-use std::env;
-use std::fs;
-use std::path::{Path, PathBuf};
+use serde::{Deserialize, Serialize};
+use std::{
+    collections::HashMap,
+    env, fs,
+    path::{Path, PathBuf},
+};
+use toml::Value;
 use walkdir::WalkDir;
 
 mod cli;
@@ -32,9 +34,9 @@ mod create_project;
 mod settings;
 mod token;
 
-#[derive(Serialize)]
-struct TagFile<'a> {
-    values: &'a Vec<String>,
+#[derive(Deserialize, Serialize)]
+struct TagFile {
+    values: Vec<String>,
 }
 
 /// Get namespace (name of folder containing the main /functions)
@@ -197,12 +199,13 @@ fn main() -> std::io::Result<()> {
 
     if datapack_is_dir {
         let mut tag_map: HashMap<String, Vec<String>> = HashMap::new();
+        let mut global_macros: HashMap<String, compiler::macros::Macro> = HashMap::new();
         let target_folder = &compiler_settings.output;
 
         if fs::metadata(target_folder).is_ok() {
             println!("Deleting old databind folder...");
             fs::remove_dir_all(&target_folder)?;
-            println!("Done.");
+            println!("Deleted.");
         }
 
         let mut inclusions = merge_globs(&compiler_settings.inclusions, datapack);
@@ -220,97 +223,204 @@ fn main() -> std::io::Result<()> {
             src_dir.as_path()
         };
 
-        for entry in WalkDir::new(&src_dir).into_iter().filter_map(|e| e.ok()) {
-            if entry.path().is_file() {
-                // Do not add config file to output folder
-                if config_path.exists() && is_same_file(entry.path(), config_path).unwrap() {
-                    continue;
-                }
-
-                let new_path_str =
-                    entry
-                        .path()
-                        .to_str()
-                        .unwrap()
-                        .replacen(src_dir.to_str().unwrap(), "", 1);
-                let path = Path::new(&new_path_str);
-
-                let mut target_path: String = target_folder.to_string();
-                target_path.push('/');
-                target_path.push_str(path.parent().unwrap().to_str().unwrap());
-
-                fs::create_dir_all(&target_path)?;
-
-                let mut compile = false;
-                let mut continue_loop = false;
-
-                for file in inclusions.iter() {
-                    if is_same_file(file, entry.path()).expect("Failed to check file paths") {
-                        compile = true;
-                        break;
-                    }
-                }
-
-                for file in exclusions.iter() {
-                    if is_same_file(file, entry.path()).expect("Failed to check file paths") {
-                        continue_loop = true;
-                        break;
-                    }
-                }
-
-                if continue_loop {
-                    continue;
-                }
-
-                if compile {
-                    let content = fs::read_to_string(entry.path())
-                        .expect(&format!("Failed to read file {}", entry.path().display())[..]);
-                    let mut compile = compiler::Compiler::new(content, true);
-                    let tokens = compile.tokenize(false);
-                    let mut compiled = compile.compile(
-                        tokens,
-                        Some(get_namespace(&entry.path()).unwrap()),
-                        &get_subfolder_prefix(&entry.path()),
-                    );
-
-                    for (key, value) in compiled.filename_map.iter() {
-                        let full_path = format!("{}/{}.mcfunction", target_path, key);
-
-                        fs::write(full_path, &compiled.file_contents[*value])?;
-
-                        // Add namespace prefix to function in tag map
-                        for (_, funcs) in compiled.tag_map.iter_mut() {
-                            if funcs.contains(key) {
-                                let i = funcs.iter().position(|x| x == key).unwrap();
-                                funcs[i] = format!(
-                                    "{}:{}{}",
-                                    get_namespace(&entry.path()).unwrap(),
-                                    get_subfolder_prefix(&entry.path()),
-                                    key
-                                );
+        // Read vars.toml if present
+        let vars = {
+            // Path to toml file
+            let path = format!("{}/vars.toml", datapack);
+            let vars_toml = Path::new(&path);
+            // Check if file exists
+            if vars_toml.exists() && vars_toml.is_file() {
+                let contents = fs::read_to_string(vars_toml)?;
+                // Read toml file into HashMap with multiple types
+                let vars_multi_type: HashMap<String, Value> = toml::from_str(&contents).unwrap();
+                let mut vars: HashMap<String, String> = HashMap::new();
+                for (k, v) in vars_multi_type.iter() {
+                    // Try to convert the value into a string
+                    let new_v: String = match v {
+                        Value::String(value) => value.clone(),
+                        Value::Boolean(value) => {
+                            if *value {
+                                "1".into()
+                            } else {
+                                "0".into()
                             }
                         }
-                    }
+                        Value::Float(value) => value.to_string(),
+                        Value::Integer(value) => value.to_string(),
+                        Value::Datetime(value) => value.to_string(),
+                        _ => {
+                            println!("[ERROR] Unsupported type found in vars.toml file (key: {}, value: {})", k, v);
+                            std::process::exit(1);
+                        }
+                    };
+                    vars.entry(format!("&{}", k)).or_insert(new_v);
+                }
+                Some(vars)
+            } else {
+                None
+            }
+        };
 
-                    tag_map.extend(compiled.tag_map);
+        // Get filepaths with global macros appearing first
+        let paths = {
+            // Store global macro filepaths
+            let mut global_macros: Vec<PathBuf> = Vec::new();
+            // Store normal filepaths
+            let mut normal: Vec<PathBuf> = Vec::new();
+            // Sort filepaths into each vector
+            let unsorted_paths = WalkDir::new(&src_dir)
+                .into_iter()
+                .filter_map(|e| e.ok())
+                .filter(|e| e.path().is_file())
+                .map(|e| e.path().to_path_buf());
+
+            for path in unsorted_paths {
+                if path.file_name().unwrap().to_str().unwrap().starts_with('!') {
+                    global_macros.push(path);
                 } else {
-                    let filename = path.file_name().unwrap().to_str().unwrap();
-                    let full_path = format!("{}/{}", target_path, filename);
-                    fs::copy(entry.path(), full_path)?;
+                    normal.push(path);
+                }
+            }
+
+            global_macros.append(&mut normal);
+            global_macros
+        };
+
+        for path in paths.iter() {
+            // Do not add config file to output folder
+            if config_path.exists() && is_same_file(path, config_path).unwrap() {
+                continue;
+            }
+
+            let new_path_str = path
+                .to_str()
+                .unwrap()
+                .replacen(src_dir.to_str().unwrap(), "", 1);
+
+            let relative_path = Path::new(&new_path_str);
+
+            let mut target_path: String = target_folder.to_string();
+            target_path.push('/');
+            target_path.push_str(relative_path.parent().unwrap().to_str().unwrap());
+
+            fs::create_dir_all(&target_path)?;
+
+            let mut compile = false;
+            let mut continue_loop = false;
+
+            for file in inclusions.iter() {
+                if is_same_file(file, path).expect("Failed to check file paths") {
+                    compile = true;
+                    break;
+                }
+            }
+
+            for file in exclusions.iter() {
+                if is_same_file(file, path).expect("Failed to check file paths") {
+                    continue_loop = true;
+                    break;
+                }
+            }
+
+            if continue_loop {
+                continue;
+            }
+
+            if compile {
+                let contents = {
+                    let mut file_contents = fs::read_to_string(path)
+                        .expect(&format!("Failed to read file {}", path.display())[..]);
+                    if let Some(vars_map) = &vars {
+                        for (k, v) in vars_map.iter() {
+                            file_contents = file_contents.replace(k, v);
+                        }
+                    }
+                    file_contents
+                };
+                let mut compile = compiler::Compiler::new(contents);
+                let tokens = compile.tokenize();
+
+                let mut compiled = if path.file_name().unwrap().to_str().unwrap().starts_with('!') {
+                    let ret = compile.compile(
+                        tokens,
+                        Some(get_namespace(&path).unwrap()),
+                        &get_subfolder_prefix(&path),
+                        &global_macros,
+                        true,
+                    );
+                    global_macros.extend(ret.global_macros.clone().unwrap());
+                    ret.clone()
+                } else {
+                    compile.compile(
+                        tokens,
+                        Some(get_namespace(&path).unwrap()),
+                        &get_subfolder_prefix(&path),
+                        &global_macros,
+                        false,
+                    )
+                };
+
+                for (key, value) in compiled.filename_map.iter() {
+                    let full_path = format!("{}/{}.mcfunction", target_path, key);
+
+                    fs::write(full_path, &compiled.file_contents[*value])?;
+
+                    // Add namespace prefix to function in tag map
+                    for (_, funcs) in compiled.tag_map.iter_mut() {
+                        if funcs.contains(key) {
+                            let i = funcs.iter().position(|x| x == key).unwrap();
+                            funcs[i] = format!(
+                                "{}:{}{}",
+                                get_namespace(&path).unwrap(),
+                                get_subfolder_prefix(&path),
+                                key
+                            );
+                        }
+                    }
+                }
+
+                tag_map.extend(compiled.tag_map);
+            } else {
+                let filename = relative_path.file_name().unwrap().to_str().unwrap();
+                let full_path = format!("{}/{}", target_path, filename);
+                // Only copy if the file doesn't exist yet
+                // Intended to stop overwriting of Databind tags
+                if fs::metadata(&full_path).is_err() {
+                    fs::copy(&path, &full_path)?;
                 }
             }
         }
 
-        // Write tag files
+        // Create tags directory
         fs::create_dir_all(format!("{}/data/minecraft/tags/functions", target_folder))?;
 
         for (tag, funcs) in tag_map.iter() {
-            let tag_file = TagFile { values: funcs };
+            let mut tag_file = TagFile {
+                values: funcs.clone(),
+            };
+
+            {
+                // Path to potential source JSON file
+                let path_str = format!(
+                    "{}/data/minecraft/tags/functions/{}.json",
+                    src_dir.display(),
+                    tag
+                );
+                let path = Path::new(&path_str);
+
+                // Read existing tags if present
+                if path.exists() && path.is_file() {
+                    let contents = fs::read_to_string(&path)?;
+                    let mut existing_tags: TagFile = serde_json::from_str(&contents)?;
+                    tag_file.values.append(&mut existing_tags.values);
+                }
+            }
+
             let json = serde_json::to_string(&tag_file)?;
 
             // Write tag file
             fs::write(
-                format!(
+                &format!(
                     "{}/data/minecraft/tags/functions/{}.json",
                     target_folder, tag
                 ),
